#include "Parser.hpp"

#include "Token.hpp"
#include "InstructionSet.hpp"
#include "Instruction.hpp"
#include "Keywords.hpp"
#include "Error.hpp"

#include <assert.h>
#include <stdio.h>
#include <string.h>

namespace SoftWire
{
	const InstructionSet Parser::instructionSet;

	Parser::Parser()
	{
	}

	Parser::~Parser()
	{
	}

	const Encoding &Parser::parseLine(const char *sourceLine)
	{
		if(!sourceLine)
		{
			throw Error::INTERNAL;
		}

		token.setSource(sourceLine);

		instruction = 0;
		synthesizer.reset();

		if(!token.isEndOfLine())
		{
			parseLabel();
		}
		if(!token.isEndOfLine())
		{
			parseMnemonic();

			if(!instruction)
			{
				throw Error("Unrecognised mnemonic");
			}

			parseFirstOperand();
			parseSecondOperand();
			parseThirdOperand();
		}

		if(instruction)
		{
			do
			{
				if(instruction->matchSyntax())
				{
					break;
				}

				instruction = instruction->getNext();
			}
			while(instruction);

			if(!instruction)
			{
				throw Error("Operands mismatch");
			}
		}

		return synthesizer.encodeInstruction(instruction);
	}

	void Parser::parseLabel()
	{
		if(token.isIdentifier() && token.lookAhead().getChar() == ':')
		{
			synthesizer.defineLabel(token.getString());

			token.advance(2);
		}
	}

	void Parser::parseMnemonic()
	{
		instruction = instructionSet.query(token.getString());

		if(instruction)
		{
			instruction->matchMnemonic(token.getString());

			token.advance();
		}
		else
		{
			throw Error("Mnemonic not recognised");
		}
	}

	void Parser::parseSpecifier()
	{
		Specifier::Type type = Specifier::UNKNOWN;

		if(token.isIdentifier())
		{
			 type = Specifier::scan(token.getString());
		}
		
		instruction->matchSpecifier(type);

		if(type != Specifier::UNKNOWN)
		{
			token.advance();	
		}
	}

	void Parser::parseFirstOperand()
	{
		assert(instruction);

		parseSpecifier();

		Operand firstOperand = Operand::INIT;

		if(token.isEndOfLine())
		{
		}
		else if(token.isPunctuator())
		{
			switch(token.getChar())
			{
			case '[':
				firstOperand = parseMemoryReference();
				break;
			case '+':
			case '-':
			case '~':
				firstOperand = parseImmediate();
				break;
			default:
				throw Error("Unexpected punctuator after mnemonic '%c'", token.getChar());
			}
		}
		else if(token.isConstant())
		{
			firstOperand = parseImmediate();
		}
		else if(token.isIdentifier())
		{
			firstOperand = parseRegister();
		}
		else
		{
			throw Error("Invalid destination operand");
		}

		instruction->matchFirstOperand(firstOperand);
		synthesizer.encodeFirstOperand(firstOperand);
	}

	void Parser::parseSecondOperand()
	{
		assert(instruction);

		if(token.getChar() == ',')
		{
			token.advance();
		}
		else if(!token.isEndOfLine())
		{
			throw Error("Operands must be separated by comma");
		}
		else
		{
			instruction->matchSecondOperand(Operand::INIT);
			return;
		}

		parseSpecifier();

		Operand secondOperand = Operand::INIT;

		if(token.isEndOfLine())
		{
		}
		else if(token.isPunctuator())
		{
			switch(token.getChar())
			{
			case '[':
				secondOperand = parseMemoryReference();
				break;
			case '+':
			case '-':
			case '~':
				secondOperand = parseImmediate();
				break;
			default:
				throw Error("Unexpected punctuator after mnemonic '%c'", token.getChar());
			}
		}
		else if(token.isConstant())
		{
			secondOperand = parseImmediate();
		}
		else if(token.isIdentifier())
		{
			secondOperand = parseRegister();
		}
		else
		{
			throw Error("Invalid source operand");
		}

		instruction->matchSecondOperand(secondOperand);
		synthesizer.encodeSecondOperand(secondOperand);
	}

	void Parser::parseThirdOperand()
	{
		assert(instruction);

		if(token.getChar() == ',')
		{
			token.advance();
		}
		else if(!token.isEndOfLine())
		{
			throw Error("Operands must be separated by comma");
		}
		else
		{
			instruction->matchThirdOperand(Operand::INIT);
			return;
		}

		Operand thirdOperand = Operand::INIT;

		if(token.isEndOfLine())
		{
		}
		else if(token.isPunctuator())
		{
			switch(token.getChar())
			{
			case '+':
			case '-':
			case '~':
				thirdOperand = parseImmediate();
				break;
			default:
				throw Error("Unexpected punctuator after mnemonic '%c'", token.getChar());
			}
		}
		else if(token.isConstant())
		{
			thirdOperand = parseImmediate();
		}
		else
		{
			throw Error("Too many operands");
		}

		instruction->matchThirdOperand(thirdOperand);
		synthesizer.encodeThirdOperand(thirdOperand);
	}

	Operand Parser::parseImmediate()
	{
		Operand imm = Operand::INIT;

		if(token.isPunctuator())
		{
			if(token.getChar() == '+')
			{
				token.advance();
				imm.value = +token.getValue();
			}
			else if(token.getChar() == '-')
			{
				token.advance();
				imm.value = -token.getValue();
			}
			else if(token.getChar() == '~')
			{
				token.advance();
				imm.value = ~token.getValue();
			}
			else
			{
				throw Error::INTERNAL;   // Method shouldn't have been called
			}
		}
		else if(token.isConstant())
		{
			imm.value = token.getValue();
		}
		else
		{
			throw Error::INTERNAL;   // Method shouldn't have been called
		}

		if((unsigned char)imm.value == imm.value)
		{
			imm.type = Operand::IMM8;
		}
		else if((unsigned short)imm.value == imm.value)
		{
			imm.type = Operand::IMM16;
		}
		else
		{
			imm.type = Operand::IMM32;
		}

		token.advance();

		return imm;
	}

	Operand Parser::parseRegister()
	{
		Operand reg = Operand::scanReg(token.getString());

		// It's not a register, so it must be a reference
		if(reg.type == Operand::UNKNOWN)
		{
			synthesizer.referenceLabel(token.getString());

			// First operand should be immediate
			reg.type = Operand::IMM8;   // Also matches IMM32
			instruction->matchFirstOperand(reg);
		}

		token.advance();

		return reg;
	}

	Operand Parser::parseMemoryReference()
	{
		Operand mem = Operand::INIT;

		while(!token.lookAhead().isEndOfLine())
		{
			const Token &prev = token.currentToken();
			token.advance();
			const Token &next = token.lookAhead();

			if(token.isIdentifier())
			{
				Operand reg = Operand::scanReg(token.getString());

				if(reg.type == Operand::UNKNOWN)
				{
					throw Error("Unrecognised identifier '%s'", token.getString());
				}

				// Check if this is a scaled index register
				if(prev.getChar() == '*' || next.getChar() == '*')
				{
					synthesizer.encodeIndex(reg);
				}
				else
				{
					synthesizer.encodeBase(reg);
				}
			}
			else if(token.isPunctuator())
			{
				switch(token.getChar())
				{
				case ']':
					mem.type = Operand::MEM;
					token.advance();
					return mem;
				case '+':
					// Check if previous and next tokens are identifier or number
					if((!prev.isConstant() && !prev.isIdentifier()) ||
					   (!next.isIdentifier() && !next.isConstant()))
					{
						throw Error("Syntax error '+' in memory reference");
					}
					break;
				case '-':
					// Check if previous and next tokens are correct type
					if((!prev.isConstant() && !prev.isIdentifier() && !(prev.getChar() == '[')) ||
					   !next.isConstant())
					{
						throw Error("Syntax error '-' in memory reference");
					}
					break;
				case '*':
					// Check if previous and next tokens are index and scale
					if(!(prev.isConstant() && next.isIdentifier()) &&
					   !(next.isConstant() && prev.isIdentifier()))
					{
						throw Error("Syntax error '*' in memory reference");
					}
					break;
				default:
					throw Error("Unexpected punctuator in memory reference '%c'", token.getChar());
				}
			}
			else if(token.isConstant())
			{
				if(prev.getChar() == '*' || next.getChar() == '*')
				{
					if(token.getValue() == 1 ||
					   token.getValue() == 2 ||
					   token.getValue() == 4 ||
					   token.getValue() == 8)
					{
						synthesizer.encodeScale(token.getValue());
					}
					else
					{
						throw Error("Invalid scale in memory reference");
					}
				}
				else if(prev.getChar() == '-')
				{
					synthesizer.encodeDisplacement(-token.getValue());
				}
				else if(prev.getChar() == '+' || next.getChar() == '+')
				{
					synthesizer.encodeDisplacement(token.getValue());
				}
				else
				{
					throw Error("Invalid number in memory reference");
				}
			}
			else
			{
				throw Error("Unexpected token in memory reference '%s'", token.getString());
			}
		}

		throw Error("Unexpected end of line in memory reference");
	}
}